VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CD3DXShape"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'  Developer:               Eric DeBrosse [MS MVP]
'                           The DirectX Cool Fool!
'                           http://www.blown.com/dx/
'
'  File:        CD3DXShape.cls
'  Content:     A class container used to create and render a D3DX shape.
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Option Explicit

Public Enum CONST_SHAPE_TYPES
    SHAPE_BOX = 0
    SHAPE_SPHERE = 1
    SHAPE_CYLINDER = 2
    SHAPE_TORUS = 3
    SHAPE_TEAPOT = 4
End Enum

'mesh and adjacency info
Dim m_oMesh As D3DXMesh
Dim m_oMeshNoColors As D3DXMesh
Dim m_MeshAdj As D3DXBuffer

'bounding box dimensions
Dim m_vBoxMin As D3DVECTOR
Dim m_vBoxMax As D3DVECTOR

'vertex types that mesh can be cloned into
Private Type D3DVERTEX_DIFFUSE
    Position As D3DVECTOR
    Normal As D3DVECTOR
    Diffuse As Long
End Type

Private Type D3DVERTEX_SPECULAR
    Position As D3DVECTOR
    Normal As D3DVECTOR
    Specular As Long
End Type

Private Type D3DVERTEX_DIFFUSE_SPECULAR
    Position As D3DVECTOR
    Normal As D3DVECTOR
    Diffuse As Long
    Specular As Long
End Type

'-----------------------------------------------------------------------------
' Name: RestoreDeviceObjects()
' Desc: Used to create the mesh.
'-----------------------------------------------------------------------------
Public Function RestoreDeviceObjects(ByVal NumSegs As Single, _
                                     Optional ByVal ShapeType As CONST_SHAPE_TYPES = SHAPE_TEAPOT) As Boolean
    On Error GoTo errout
    
    Select Case ShapeType
        Case SHAPE_BOX
            Set m_oMeshNoColors = g_d3dx.CreateBox(g_dev, 1!, 1!, 1!, m_MeshAdj)
            Tessellate NumSegs + 1
            
        Case SHAPE_SPHERE
            Set m_oMeshNoColors = g_d3dx.CreateSphere(g_dev, 1!, Detail(NumSegs), Detail(NumSegs), m_MeshAdj)
            
        Case SHAPE_CYLINDER
            Set m_oMeshNoColors = g_d3dx.CreateCylinder(g_dev, 0.7!, 1!, 1!, Detail(NumSegs), Detail(NumSegs), m_MeshAdj)
            
        Case SHAPE_TORUS
            Set m_oMeshNoColors = g_d3dx.CreateTorus(g_dev, 0.7!, 1!, Detail(NumSegs), Detail(NumSegs), m_MeshAdj)
            
        Case SHAPE_TEAPOT
            Set m_oMeshNoColors = g_d3dx.CreateTeapot(g_dev, m_MeshAdj)
            If NumSegs > 5 Then
                'limit the tessellation, the teapot can really generate a lot of faces
                Tessellate 6
            Else
                Tessellate NumSegs + 1
            End If
    End Select
    
    Set m_oMesh = m_oMeshNoColors
    
    'update the bounding box
    g_d3dx.ComputeBoundingBoxFromMesh m_oMesh, m_vBoxMin, m_vBoxMax
    
    RestoreDeviceObjects = True
Exit Function

errout:
    Debug.Print "Failed to create mesh."
    RestoreDeviceObjects = False
End Function

'-----------------------------------------------------------------------------
' Name: Detail()
' Desc:
'-----------------------------------------------------------------------------
Private Function Detail(ByVal ValueIn As Long) As Long
    Detail = 5 + ValueIn * 10
End Function

'-----------------------------------------------------------------------------
' Name: InvalidateDeviceObjects()
' Desc: The mesh must be destroyed before a device reset.
'-----------------------------------------------------------------------------
Public Sub InvalidateDeviceObjects()
    On Error Resume Next
    
    Set m_oMeshNoColors = Nothing
    Set m_oMesh = Nothing
End Sub

'-----------------------------------------------------------------------------
' Name: Render()
' Desc: Used to render the mesh.
'-----------------------------------------------------------------------------
Public Function Render() As Boolean
    On Error GoTo errout
    
    m_oMesh.DrawSubset 0&
    
    Render = True
Exit Function

errout:
    Debug.Print "Failed to render mesh."
    Render = False
End Function

'-----------------------------------------------------------------------------
' Name: Optimize()
' Desc: Used to optimize the mesh.
'-----------------------------------------------------------------------------
Public Function Optimize() As Boolean
    Dim BufSize As Long
    Dim AdjacencyOut() As Long
    Dim FaceRemap() As Long
    Dim VertexRemapOut As D3DXBuffer
    
    On Error GoTo errout
    
    'make room for enough Long values to represent the mesh data
    BufSize = m_MeshAdj.GetBufferSize
    ReDim AdjacencyOut(BufSize / 4&)
    ReDim FaceRemap(BufSize / 4&)
    
    'optimize mesh data
    m_oMeshNoColors.OptimizeInplace D3DXMESHOPT_ATTRSORT, ByVal m_MeshAdj.GetBufferPointer, _
            AdjacencyOut(0), FaceRemap(0), VertexRemapOut
    
    'update copy of mesh
    Set m_oMesh = Nothing
    Set m_oMesh = m_oMeshNoColors
    
    Optimize = True
Exit Function

errout:
    Debug.Print "Failed to optimize mesh."
    Optimize = False
End Function

'-----------------------------------------------------------------------------
' Name: CloneFVF()
' Desc: Clone the mesh into a mesh with the requested
'       colors for each vertex.  Then compute the requested
'       colors into the cloned data.
'-----------------------------------------------------------------------------
Public Function CloneFVF(ByVal fvf As CONST_D3DFVFFLAGS) As Boolean
    Dim i As Long
    Dim LockedData As Long
    Dim LockedArray As Long
    
    On Error GoTo errout
    
    'destroy current copy of mesh
    Set m_oMesh = Nothing
    
    'fvf must only contain specular, diffuse, or both
    If fvf <> D3DFVF_SPECULAR And fvf <> D3DFVF_DIFFUSE And _
            fvf <> (D3DFVF_DIFFUSE Or D3DFVF_SPECULAR) Then
        Set m_oMesh = m_oMeshNoColors
        CloneFVF = False
        Exit Function
    End If
    
    'clone mesh based on specified fvf
    Set m_oMesh = m_oMeshNoColors.CloneMeshFVF(D3DXMESH_MANAGED, _
            (D3DFVF_XYZ Or D3DFVF_NORMAL Or fvf), g_dev)
    
    'compute the requested colors into the cloned data
    If fvf = D3DFVF_DIFFUSE Then
        Dim MyVerts() As D3DVERTEX_DIFFUSE
        ReDim MyVerts(m_oMesh.GetNumVertices)
        
        LockedData = m_oMesh.LockVertexBuffer(0)
        LockedArray = DXLockArray8(m_oMesh.GetVertexBuffer, LockedData, MyVerts)
        
        For i = 0 To m_oMesh.GetNumVertices - 1
            MyVerts(i).Diffuse = ComputeDiffuseColor(MyVerts(i).Position)
        Next i
        
        LockedArray = DXUnlockArray8(m_oMesh.GetVertexBuffer, MyVerts)
        m_oMesh.UnlockVertexBuffer
        
    ElseIf fvf = D3DFVF_SPECULAR Then
        Dim MyVerts2() As D3DVERTEX_SPECULAR
        ReDim MyVerts2(m_oMesh.GetNumVertices)
        
        LockedData = m_oMesh.LockVertexBuffer(0)
        LockedArray = DXLockArray8(m_oMesh.GetVertexBuffer, LockedData, MyVerts2)
        
        For i = 0 To m_oMesh.GetNumVertices - 1
            MyVerts2(i).Specular = ComputeSpecularColor(MyVerts2(i).Position, m_vBoxMin, m_vBoxMax)
        Next i
        
        LockedArray = DXUnlockArray8(m_oMesh.GetVertexBuffer, MyVerts2)
        m_oMesh.UnlockVertexBuffer
        
    Else 'D3DFVF_DIFFUSE Or D3DFVF_SPECULAR
        Dim MyVerts3() As D3DVERTEX_DIFFUSE_SPECULAR
        ReDim MyVerts3(m_oMesh.GetNumVertices)
        
        LockedData = m_oMesh.LockVertexBuffer(0)
        LockedArray = DXLockArray8(m_oMesh.GetVertexBuffer, LockedData, MyVerts3)
        
        For i = 0 To m_oMesh.GetNumVertices - 1
            MyVerts3(i).Diffuse = ComputeDiffuseColor(MyVerts3(i).Position)
            MyVerts3(i).Specular = ComputeSpecularColor(MyVerts3(i).Position, m_vBoxMin, m_vBoxMax)
        Next i
        
        LockedArray = DXUnlockArray8(m_oMesh.GetVertexBuffer, MyVerts3)
        m_oMesh.UnlockVertexBuffer
    End If
    
    CloneFVF = True
Exit Function

errout:
    Debug.Print "Failed to clone mesh."
    CloneFVF = False
End Function

'-----------------------------------------------------------------------------
' Name: Tessellate()
' Desc: Used to tessellate the mesh using the N-patch
'       algorithm to get smoother lighting than with the
'       plain mesh.
'-----------------------------------------------------------------------------
Private Function Tessellate(ByVal NumSegs As Single, _
                            Optional ByVal QuadraticInterpNormals As Boolean = True) As Boolean
    Dim oAdjOut As D3DXBuffer
    Dim oTempMesh As D3DXMesh
    
    On Error GoTo errout
    
    If NumSegs <= 0 Then
        'no tessellation needed
        Tessellate = False
        Exit Function
    End If
    
    If m_oMeshNoColors Is Nothing Then
        Tessellate = False
        Exit Function
    End If
    
    'tessellate the mesh
    Set oTempMesh = g_d3dx.TessellateNPatches(m_oMeshNoColors, _
            ByVal m_MeshAdj.GetBufferPointer, NumSegs, QuadraticInterpNormals, oAdjOut)
    
    'mesh was created, destroy old mesh
    Set m_oMeshNoColors = Nothing
    Set m_oMeshNoColors = oTempMesh
    
    Set m_MeshAdj = Nothing
    Set m_MeshAdj = oAdjOut
    
    'update copy of mesh
    Set m_oMesh = Nothing
    Set m_oMesh = m_oMeshNoColors
    
    Tessellate = True
Exit Function

errout:
    Debug.Print "Failed to tessellate mesh."
    Tessellate = False
End Function

'-----------------------------------------------------------------------------
' Name: ComputeDiffuseColor()
' Desc: Compute a diffuse color for the mesh by creating a hue
'       ramp in the x direction.
'-----------------------------------------------------------------------------
Private Function ComputeDiffuseColor(ByRef vPosition As D3DVECTOR) As Long
    Dim Diffuse As D3DCOLORVALUE
    Dim Hue As Single
    
    Hue = (vPosition.x + 1!) * 360!
    hsv_to_rgb Hue, 1!, 1!, Diffuse.r, Diffuse.g, Diffuse.b
    
    ComputeDiffuseColor = D3DCOLORAUX.D3DColorMake(Diffuse.r, Diffuse.g, Diffuse.b, 1!)
End Function

'-----------------------------------------------------------------------------
' Name: ComputeSpecularColor()
' Desc: Compute a specular color by creating a grayscale gradient
'       over the top half of the mesh.
'-----------------------------------------------------------------------------
Private Function ComputeSpecularColor(ByRef vPosition As D3DVECTOR, _
                                      ByRef vMinima As D3DVECTOR, _
                                      ByRef vMaxima As D3DVECTOR) As Long
    Dim fFraction As Single
    Dim fTemp As Single
    Dim Intensity As Byte
    
    fFraction = (vPosition.y - vMinima.y) / (vMaxima.y - vMinima.y)
    
    If fFraction > 0.5! Then
        fTemp = (fFraction - 0.5!) * 2!
        If fTemp < 0! Then fTemp = 0!
        If fTemp > 1! Then fTemp = 1!
        
        Intensity = CByte(255 * fTemp)
        
    Else
        Intensity = 0
    End If
    
    ComputeSpecularColor = D3DCOLORAUX.D3DColorXRGB(Intensity, Intensity, Intensity)
End Function

'-----------------------------------------------------------------------------
' Name: hsv_to_rgb()
' Desc: Convert hue, saturation, value to red, green, blue.
'-----------------------------------------------------------------------------
Private Sub hsv_to_rgb(ByVal h As Single, _
                       ByVal s As Single, _
                       ByVal v As Single, _
                       ByRef r As Single, _
                       ByRef g As Single, _
                       ByRef b As Single)
    Dim f As Single
    Dim h2 As Single
    Dim p1 As Single
    Dim p2 As Single
    Dim p3 As Single
    
    If s < 0! Or s > 1! Then
        Debug.Print "hsv_to_rgb - saturation parameter out of range: " & CStr(s)
        Exit Sub
    End If
    
    If v < 0! Or v > 1! Then
        Debug.Print "hsv_to_rgb - value parameter out of range: " & CStr(v)
        Exit Sub
    End If
    
    h2 = h
    
    If h2 < 0! Then
        h2 = h2 - Fix((h2 / 360!) - 1!) * 360!
    ElseIf h2 >= 360! Then
        h2 = h2 - Fix(h2 / 360!) * 360!
    End If
    
    h2 = h2 / 60!
    f = h2 - Fix(h2)
    
    p1 = v * (1! - s)
    p2 = v * (1! - (s * f))
    p3 = v * (1! - (s * (1! - f)))
    
    Select Case Fix(h2)
        Case 0
            r = v: g = p3: b = p1
        Case 1
            r = p2: g = v:  b = p1
        Case 2
            r = p1: g = v:  b = p3
        Case 3
            r = p1: g = p2: b = v
        Case 4
            r = p3: g = p1: b = v
        Case 5
            r = v:  g = p1: b = p2
    End Select
End Sub

